# cython: profile=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: warn.undeclared=True
cimport openTSNE._matrix_mul.matrix_mul
cimport numpy as np
import numpy as np


cdef double[:, ::1] matrix_multiply_fft_1d(
    double[::1] kernel_tilde,
    double[:, ::1] w_coefficients,
):
    """Multiply the the kernel vectr K tilde with the w coefficients.
    
    Parameters
    ----------
    kernel_tilde : memoryview
        The generating vector of the 2d Toeplitz matrix i.e. the kernel 
        evaluated all all interpolation points from the left most 
        interpolation point, embedded in a circulant matrix (doubled in size 
        from (n_interp, n_interp) to (2 * n_interp, 2 * n_interp) and 
        symmetrized. See how to embed Toeplitz into circulant matrices.
    w_coefficients : memoryview
        The coefficients calculated in Step 1 of the paper, a
        (n_total_interp, n_terms) matrix. The coefficients are embedded into a
        larger matrix in this function, so no prior embedding is needed.
        
    Returns
    -------
    memoryview
        Contains the kernel values at the equspaced interpolation nodes.
    
    """
    cdef:
        Py_ssize_t n_interpolation_points_1d = w_coefficients.shape[0]
        Py_ssize_t n_terms = w_coefficients.shape[1]
        Py_ssize_t n_fft_coeffs = kernel_tilde.shape[0]

        double[:, ::1] y_tilde_values = np.empty((n_interpolation_points_1d, n_terms), dtype=float)

        complex[::1] fft_kernel_tilde = np.empty(n_fft_coeffs, dtype=complex)
        complex[::1] fft_w_coeffs = np.empty(n_fft_coeffs, dtype=complex)
        # Note that we can't use the same buffer for the input and output since
        # we only write to the first half of the vector - we'd need to
        # manually zero out the rest of the entries that were inevitably
        # changed during the IDFT, so it's faster to use two buffers, at the
        # cost of some memory
        complex[::1] fft_in_buffer = np.zeros(n_fft_coeffs, dtype=complex)
        complex[::1] fft_out_buffer = np.zeros(n_fft_coeffs, dtype=complex)

        Py_ssize_t d, i

    # Compute the FFT of the kernel vector
    fft_kernel_tilde = np.fft.fft(kernel_tilde)

    for d in range(n_terms):
        for i in range(n_interpolation_points_1d):
            fft_in_buffer[i] = w_coefficients[i, d]

        fft_w_coeffs = np.fft.fft(fft_in_buffer)

        # Take the Hadamard product of two complex vectors
        fft_w_coeffs = np.multiply(fft_w_coeffs, fft_kernel_tilde)

        fft_out_buffer = np.fft.ifft(fft_w_coeffs)

        for i in range(n_interpolation_points_1d):
            y_tilde_values[i, d] = fft_out_buffer[n_interpolation_points_1d + i].real

    return y_tilde_values


cdef double[:, ::1] matrix_multiply_fft_2d(
    double[:, ::1] kernel_tilde,
    double[:, ::1] w_coefficients,
):
    """Multiply the the kernel matrix K tilde with the w coefficients.

    Parameters
    ----------
    kernel_tilde : memoryview
        The generating matrix of the 3d Toeplitz tensor i.e. the kernel
        evaluated all all interpolation points from the top left most
        interpolation point, embedded in a circulant matrix (doubled in size
        from (n_interp, n_interp) to (2 * n_interp, 2 * n_interp) and
        symmetrized. See how to embed Toeplitz into circulant matrices.
    w_coefficients : memoryview
        The coefficients calculated in Step 1 of the paper, a
        (n_total_interp, n_terms) matrix. The coefficients are embedded into a
        larger matrix in this function, so no prior embedding is needed.

    Returns
    -------
    memoryview
        Contains the kernel values at the equspaced interpolation nodes.

    """
    cdef:
        Py_ssize_t total_interpolation_points = w_coefficients.shape[0]
        Py_ssize_t n_terms = w_coefficients.shape[1]
        Py_ssize_t n_fft_coeffs = kernel_tilde.shape[0]
        Py_ssize_t n_interpolation_points_1d = n_fft_coeffs / 2

        double[:, ::1] y_tilde_values = np.empty((total_interpolation_points, n_terms))

        complex[:, :] fft_w_coefficients = np.empty((n_fft_coeffs, (n_fft_coeffs / 2 + 1)), dtype=complex)
        complex[:, :] fft_kernel_tilde = np.empty((n_fft_coeffs, (n_fft_coeffs / 2 + 1)), dtype=complex)
        # Note that we can't use the same buffer for the input and output since
        # we only write to the top quadrant of the in matrix - we'd need to
        # manually zero out the rest of the entries that were inevitably
        # changed during the IDFT, so it's faster to use two buffers, at the
        # cost of some memory
        double[:, ::1] fft_in_buffer = np.zeros((n_fft_coeffs, n_fft_coeffs), dtype=float)
        double[:, ::1] fft_out_buffer = np.zeros((n_fft_coeffs, n_fft_coeffs), dtype=float)

        Py_ssize_t d, i, j, idx

    fft_kernel_tilde = np.fft.rfft2(kernel_tilde)

    for d in range(n_terms):
        for i in range(n_interpolation_points_1d):
            for j in range(n_interpolation_points_1d):
                fft_in_buffer[i, j] = w_coefficients[i * n_interpolation_points_1d + j, d]

        fft_w_coefficients = np.fft.rfft2(fft_in_buffer)

        # Take the Hadamard product of two complex vectors
        fft_w_coefficients = np.multiply(fft_w_coefficients, fft_kernel_tilde)

        # Invert the computed values at the interpolated nodes
        fft_out_buffer = np.fft.irfft2(fft_w_coefficients)

        for i in range(n_interpolation_points_1d):
            for j in range(n_interpolation_points_1d):
                idx = i * n_interpolation_points_1d + j
                y_tilde_values[idx, d] = fft_out_buffer[n_interpolation_points_1d + i,
                                                        n_interpolation_points_1d + j].real

    return y_tilde_values



# cython: profile=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: warn.undeclared=True
from libcpp.vector cimport vector
from cython.parallel import prange
import numpy as np
cimport numpy as np

cdef extern from *:
    ctypedef int euclidean_distance_t "euclidean_distance"

cdef extern from "vptree.h":
    cdef cppclass DataPoint:
        DataPoint() nogil except +
        DataPoint(int, int, double*) nogil except +
        DataPoint(DataPoint&) nogil
        int index() nogil
        int dimensionality() nogil

    cdef cppclass VpTree[T, F]:
        VpTree() nogil except +
        void create(vector[T]&) nogil
        void search(T&, int, vector[T]*, vector[double]*) nogil


cdef class VPTree:
    cdef VpTree[DataPoint, euclidean_distance_t]* tree
    cdef vector[DataPoint] train_data
    cdef double[:, ::1] data

    valid_metrics = ["euclidean"]

    def __cinit__(self):
        self.tree = new VpTree[DataPoint, euclidean_distance_t]()

    def __dealloc__(self):
        del self.tree

    def __init__(self, double[:, ::1] data):
        # It is important to keep a reference to ``data`` because otherwise
        # (this is an educated guess), the ``*data`` field in DataPoint is
        # dereferenced during execution, resulting in segfaults.
        self.data = data
        self.create(data)

    cdef create(self, double[:, ::1] data):
        # Put the data into a vector of ``DataPoint`` instances
        self.train_data.clear()
        cdef Py_ssize_t idx, N = data.shape[0], n_dim = data.shape[1]

        for idx in range(N):
            self.train_data.push_back(DataPoint(n_dim, idx, &data[idx, 0]))

        self.tree.create(self.train_data)

    def query_train(self, int k, Py_ssize_t num_threads=1):
        # Define objects to be returned
        cdef Py_ssize_t N = self.train_data.size()
        cdef np.int64_t[:, ::1] indices = np.empty((N, k), dtype=np.int64)
        cdef double[:, ::1] distances = np.empty((N, k), dtype=np.float64)

        self.__query(self.train_data, k, indices, distances, num_threads)

        return np.asarray(indices, dtype=np.int64), np.asarray(distances, np.float64)

    def query(self, double[:, ::1] query, int k, Py_ssize_t num_threads=1):
        cdef vector[DataPoint] query_points
        cdef Py_ssize_t idx, N = query.shape[0], n_dim = query.shape[1]

        # Define objects to be returned
        cdef np.int64_t[:, ::1] indices = np.empty((N, k), dtype=np.int64)
        cdef double[:, ::1] distances = np.empty((N, k), dtype=np.float64)

        for idx in range(N):
            query_points.push_back(DataPoint(n_dim, idx, &query[idx, 0]))

        self.__query(query_points, k, indices, distances, num_threads)

        return np.asarray(indices, dtype=np.int64), np.asarray(distances, np.float64)

    cdef void __query(
            self,
            vector[DataPoint] query,
            int k,
            np.int64_t[:, ::1] indices,
            double[:, ::1] distances,
            Py_ssize_t num_threads=1,
    ):
        cdef Py_ssize_t i, j, N = query.size(), n_dim = query.at(0).dimensionality()
        # Define objects to be used internally in vptree
        cdef vector[DataPoint]* indices_
        cdef vector[double]* distances_

        if num_threads < 1:
            num_threads = 1

        for i in prange(N, nogil=True, schedule="guided", num_threads=num_threads):
            indices_ = new vector[DataPoint]()
            distances_ = new vector[double]()

            self.tree.search(query.at(i), k, indices_, distances_)

            for j in range(k):
                indices[i, j] = indices_.at(j).index()
                distances[i, j] = distances_.at(j)

            del indices_, distances_



import os
import tempfile
import warnings
from distutils import ccompiler
from distutils.command.build_ext import build_ext
from distutils.errors import CompileError, LinkError
from os.path import join

import setuptools
from setuptools import setup, Extension

try:
    import numpy
except ImportError:
    from subprocess import call
    call(["pip", "install", "numpy"])

import numpy as np


def has_c_library(library, extension=".c"):
    """Check whether a C/C++ library is available on the system to the compiler.

    Parameters
    ----------
    library: str
        The library we want to check for e.g. if we are interested in FFTW3, we
        want to check for `fftw3.h`, so this parameter will be `fftw3`.
    extension: str
        If we want to check for a C library, the extension is `.c`, for C++
        `.cc`, `.cpp` or `.cxx` are accepted.

    Returns
    -------
    bool
        Whether or not the library is available.

    """
    with tempfile.TemporaryDirectory(dir=".") as directory:
        name = join(directory, "%s%s" % (library, extension))
        with open(name, "w") as f:
            f.write("#include <%s.h>\n" % library)
            f.write("int main() {}\n")

        # Get a compiler instance
        compiler = ccompiler.new_compiler()
        assert isinstance(compiler, ccompiler.CCompiler)

        try:
            # Try to compile the file using the C compiler
            compiler.link_executable(compiler.compile([name]), name)
            return True
        except (CompileError, LinkError):
            return False


class CythonBuildExt(build_ext):

    COMPILER_FLAGS = {
        "unix": {"openmp": "-fopenmp",
                 "optimize": "-O3",
                 "fftw": "-lfftw3",
                 "math": "-lm",
                 },
        "msvc": {"openmp": "/openmp",
                 "optimize": "/Ox",
                 "fftw": "/lfftw3",
                 "math": "/lm"
                 },
    }

    def build_extensions(self):
        # Automatically append the file extension based on language.
        # ``cythonize`` does this for us automatically, so it's not necessary if
        # that was run
        if not HAS_CYTHON:
            for extension in self.extensions:
                for idx, source in enumerate(extension.sources):
                    base, ext = os.path.splitext(source)
                    base += ".cpp" if extension.language == "c++" else ".c"
                    extension.sources[idx] = base

        # Optimization compiler/linker flags are added appropriately
        flags = self.COMPILER_FLAGS[self.compiler.compiler_type]
        compile_flags = [flags["optimize"]]
        link_flags = [flags["optimize"]]

        # Map any existing compile/link flags into compiler specific ones
        def map_flags(ls):
            return list(map(lambda flag: flags.get(flag, flag), ls))

        for extension in self.extensions:
            extension.extra_compile_args = map_flags(extension.extra_compile_args)
            extension.extra_link_args = map_flags(extension.extra_link_args)

        # We will disable openmp flags if the compiler doesn"t support it. This
        # is only really an issue with OSX clang
        if has_c_library("omp"):
            print("Found openmp. Compiling with openmp flags...")
            compile_flags.append(flags["openmp"]), link_flags.append(flags["openmp"])
        else:
            warnings.warn(
                "You appear to be using a compiler which does not support "
                "openMP, meaning that the library will not be able to run on "
                "multiple cores. Please install/enable openMP to use multiple "
                "cores."
            )

        for extension in self.extensions:
            extension.extra_compile_args.extend(compile_flags)
            extension.extra_link_args.extend(link_flags)

        # We typically use numpy includes in our Cython files, so we"ll add the
        # appropriate headers here
        for extension in self.extensions:
            extension.include_dirs.append(np.get_include())

        super().build_extensions()


extensions = [
    Extension("openTSNE.vptree", ["openTSNE/vptree.pyx"], language="c++"),
    Extension("openTSNE.quad_tree", ["openTSNE/quad_tree.pyx"]),
    Extension("openTSNE._tsne", ["openTSNE/_tsne.pyx"]),
    Extension("openTSNE.kl_divergence", ["openTSNE/kl_divergence.pyx"]),
]

# Check if we have access to FFTW3 and if so, use that implementation
if has_c_library("fftw3"):
    print("FFTW3 header files found. Using FFTW implementation of FFT.")
    extensions.append(
        Extension("openTSNE._matrix_mul.matrix_mul",
                  ["openTSNE/_matrix_mul/matrix_mul_fftw3.pyx"],
                  extra_compile_args=["fftw", "math"],
                  extra_link_args=["fftw", "math"],
                  )
        )
else:
    print("FFTW3 header files not found. Using numpy implementation of FFT.")
    extensions.append(
        Extension("openTSNE._matrix_mul.matrix_mul",
                  ["openTSNE/_matrix_mul/matrix_mul_numpy.pyx"])
    )

try:
    from Cython.Build import cythonize
    extensions = cythonize(extensions)
    HAS_CYTHON = True
except ImportError:
    HAS_CYTHON = False

setup(
    name="openTSNE",
    description="",
    version="0.3.0",
    license="BSD-3-Clause",

    author="Pavlin Poličar",
    author_email="pavlin.g.p@gmail.com",
    url="https://github.com/pavlin-policar/openTSNE",
    project_urls={
        "Documentation": "https://opentsne.readthedocs.io/",
        "Source": "https://github.com/pavlin-policar/openTSNE",
        "Issue Tracker": "https://github.com/pavlin-policar/openTSNE/issues",
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "License :: OSI Approved",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],

    packages=setuptools.find_packages(include=["openTSNE", "openTSNE.*"]),
    install_requires=[
        "numpy>1.14",
        "numba>=0.38.1",
        "scikit-learn>=0.20",
        "scipy",
    ],

    ext_modules=extensions,
    cmdclass={"build_ext": CythonBuildExt},
)
